import html
import importlib
import importlib.util
import os
import traceback

from concurrent.futures import Future, ThreadPoolExecutor

import panel as pn
import param
import yaml

from panel.template.base import BasicTemplate

from .auth import AuthPlugin
from .base import Component
from .config import (
    _DEFAULT_LAYOUT, _LAYOUTS, _TEMPLATES, _THEMES, config,
)
from .filters import ConstantFilter, Filter, WidgetFilter  # noqa
from .panel import IconButton
from .pipeline import Pipeline
from .sources import RESTSource, Source  # noqa
from .state import state
from .target import Target
from .transforms import Transform  # noqa
from .util import catch_and_notify, expand_spec, resolve_module_reference
from .validation import ValidationError, match_suggestion_message
from .variables import Variable, Variables
from .views import View  # noqa

pn.config.css_files.append(
    'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.3/css/all.min.css'
)

def load_yaml(yaml_spec, **kwargs):
    expanded = expand_spec(yaml_spec, config.template_vars, **kwargs)
    return yaml.load(expanded, Loader=yaml.Loader)


class Config(Component):
    """
    High-level configuration options for the Dashboard.
    """

    auto_update = param.Boolean(default=True, constant=True, doc="""
        Whether changes in filters, transforms and references automatically
        trigger updates in the data or whether an update has to be triggered
        manually using the update event or the update button in the UI."""
    )

    background_load = param.Boolean(default=False, constant=True, doc="""
        Whether to load any targets in the background.""")

    editable = param.Boolean(default=False, constant=True, doc="""
        Whether the dashboard specification is editable from within
        the deployed dashboard.""")

    layout = param.Selector(default=_DEFAULT_LAYOUT, constant=True, objects=_LAYOUTS,
                            check_on_set=False, doc="""
        Overall layout of the dashboard.""")

    loading_spinner = param.Selector(default='dots', constant=True, objects=[
        'arc', 'arcs', 'bar', 'dots', 'petal'], doc="""
        Loading indicator to use when component loading parameter is set.""")

    loading_color = param.Color(default='#00aa41', constant=True, doc="""
        Color of the loading indicator.""")

    logo = param.String(default=None, constant=True, doc="""
        A logo to add to the theme.""")

    ncols = param.Integer(default=3, bounds=(1, None), constant=True, doc="""
        Number of columns to lay out targets in.""")

    reloadable = param.Boolean(default=True, constant=True, doc="""
        Whether to allow reloading data from source(s) using a button.""")

    sync_with_url = param.Boolean(default=False, constant=True, doc="""
        Whether to sync current state of the application.""")

    title = param.String(default="Lumen Dashboard", constant=True, doc="""
        The title of the dashboard.""")

    template = param.Selector(default=_TEMPLATES['material'], constant=True, objects=_TEMPLATES,
                              check_on_set=False, doc="""
        The Panel template to render the dashboard into.""")

    theme = param.Selector(default=_THEMES['default'], objects=_THEMES, constant=True,
                           check_on_set=False, doc="""
        The Panel template theme to style the dashboard with.""")

    _valid_keys = 'params'
    _validate_params = True

    @classmethod
    def _validate_layout(cls, layout, spec, context):
        if layout not in _LAYOUTS:
            msg = f'Config layout {layout!r} could not be found. Layout must be one of {list(_LAYOUTS)}.'
            raise ValidationError(msg, spec, 'layout')
        return layout

    @classmethod
    def _validate_template(cls, template, spec, context):
        if template in _TEMPLATES:
            return template
        elif '.' not in template:
            raise ValidationError(
                f'Config template {template!r} not found. Template must be one '
                'of {list(_TEMPLATES)} or an absolute import path.', spec, template
            )
        *paths, name = template.split('.')
        path = '.'.join(paths)
        try:
            module = importlib.import_module(path)
        except Exception as e:
            raise ValidationError(
                f'Config template {path!r} module could not be imported '
                f'errored with: {e}.', spec, path
            )
        if not hasattr(module, name):
            raise ValidationError(
                f'Config template {name!r} was not found in {path!r} module.',
                spec, name
            )
        template_cls = getattr(module, name)
        if not issubclass(template_cls, BasicTemplate):
            raise ValidationError(
                f'Config template \'{path}.{name}\' is not a valid Panel template.',
                spec, 'template'
            )
        return template

    @classmethod
    def _validate_theme(cls, theme, spec, context):
        if theme not in _THEMES:
            msg = f'Config theme {theme!r} could not be found. Theme must be one of {list(_THEMES)}.'
            raise ValidationError(msg, spec, 'theme')
        return theme

    @classmethod
    def from_spec(cls, spec):
        if 'template' in spec:
            template = spec['template']
            if template in _TEMPLATES:
                template_cls = _TEMPLATES[template]
            else:
                template_cls = resolve_module_reference(template, BasicTemplate)
            spec['template'] = template_cls
        if 'theme' in spec:
            spec['theme'] = _THEMES[spec['theme']]
        if 'layout' in spec:
            spec['layout'] = _LAYOUTS[spec['layout']]
        return cls(**spec)

    def to_spec(self, context=None):
        spec = super().to_spec(context=context)
        if 'layout' in spec:
            spec['layout'] = {v: k for k, v in _LAYOUTS.items()}[spec['layout']]
        if 'template' in spec:
            tmpl = spec['template']
            spec['template'] = {v: k for k, v in _TEMPLATES.items()}.get(tmpl, f'{tmpl.__module__}.{tmpl.__name__}')
        if 'theme' in spec:
            spec['theme'] = {v: k for k, v in _THEMES.items()}[spec['theme']]
        return spec

    def __init__(self, **params):
        super().__init__(**params)
        pn.config.param.update(
            loading_spinner=self.loading_spinner,
            loading_color=self.loading_color
        )

    def construct_template(self):
        params = {'title': self.title, 'theme': self.theme}
        if self.logo:
            params['logo'] = self.config.logo
        return self.template(**params)


class Defaults(Component):
    """
    Defaults to apply to the component classes.
    """

    filters = param.List(doc="Defaults for Filter objects.", class_=dict)

    sources = param.List(doc="Defaults for Source objects.", class_=dict)

    transforms = param.List(doc="Defaults for Transform objects.", class_=dict)

    views = param.List(doc="Defaults for View objects.", class_=dict)

    _valid_keys = 'params'
    _validate_params = True

    @classmethod
    def _validate_defaults(cls, defaults_type, defaults, spec):
        highlight = f'{defaults_type.__name__.lower()}s'
        if not isinstance(defaults, list):
            raise ValidationError(
                f'Defaults must be defined as a list not as {type(defaults)}.',
                spec, highlight
            )
        for default in defaults:
            if 'type' not in default:
                raise ValidationError(
                    'Defaults must declare the type the defaults apply to.',
                    spec, highlight
                )
            default = dict(default)
            dtype = default.pop('type')
            obj_type = defaults_type._get_type(dtype)
            for p in default:
                if p in obj_type.param:
                    pobj = obj_type.param[p]
                    try:
                        pobj._validate(default[p])
                    except Exception as e:
                        msg = f"The default for {obj_type.__name__} {p!r} parameter failed validation: {str(e)}"
                        raise ValidationError(msg, default, p)
                else:
                    msg = (
                        f'Default for {obj_type.__name__} {p!r} parameter cannot be set '
                        'as there is no such parameter.'
                    )
                    msg = match_suggestion_message(p, list(obj_type.param), msg)
                    raise ValidationError(msg, default, p)
        return defaults

    @classmethod
    def _validate_filters(cls, filter_defaults, spec, context):
        return cls._validate_defaults(Filter, filter_defaults, spec)

    @classmethod
    def _validate_sources(cls, source_defaults, spec, context):
        return cls._validate_defaults(Source, source_defaults, spec)

    @classmethod
    def _validate_transforms(cls, transform_defaults, spec, context):
        return cls._validate_defaults(Transform, transform_defaults, spec)

    @classmethod
    def _validate_views(cls, view_defaults, spec, context):
        return cls._validate_defaults(View, view_defaults, spec)

    def apply(self):
        for obj, defaults in ((Filter, self.filters), (Source, self.sources),
                              (Transform, self.transforms), (View, self.views)):
            for default in defaults:
                params = dict(default)
                obj_type = obj._get_type(params.pop('type'))
                obj_type.param.set_param(**params)


class Auth(Component):

    case_sensitive = param.Boolean(default=False, constant=True, doc="""
        Whether auth validation is case-sensitive or not.""")

    spec = param.Dict({}, constant=True, doc="""
        Dictionary of keys and values to match the pn.state.user_info
        against.""")

    _allows_refs = False

    @classmethod
    def from_spec(cls, spec):
        spec = dict(spec)
        case_sensitive = spec.pop('case_sensitive', cls.case_sensitive)
        plugins = spec.pop('plugins', [])
        for plugin_spec in plugins:
            plugin = AuthPlugin.from_spec(plugin_spec)
            spec = plugin.transform(spec)
        return cls(spec=spec, case_sensitive=case_sensitive)

    @property
    def authorized(self):
        if not self.spec:
            return True
        elif pn.state.user_info is None and self.spec:
            return config.dev
        authorized = True
        for k, value in self.spec.items():
            if not isinstance(value, list): value = [value]
            if not self.case_sensitive:
                value = [v.lower() if isinstance(v, str) else v for v in value]
            value = [v.strip() if isinstance(v, str) else v for v in value]
            if k in pn.state.user_info:
                user_value = pn.state.user_info[k]
                if not isinstance(user_value, list):
                    user_value = [user_value]
                if not self.case_sensitive:
                    user_value = [
                        uv.lower() if isinstance(uv, str) else uv
                        for uv in user_value
                    ]
                authorized &= any(uv == v for v in value for uv in user_value)
        return authorized


class Dashboard(Component):

    auth = param.ClassSelector(default=Auth(), class_=Auth, doc="""
        Auth object which validates the auth spec against pn.state.user_info.""")

    config = param.ClassSelector(default=Config(), class_=Config, doc="""
        High-level configuration options for the dashboard.""")

    defaults = param.ClassSelector(default=Defaults(), class_=Defaults, doc="""
        Defaults to apply to component classes.""")

    targets = param.List(default=[], item_type=Target, doc="""
        List of targets monitoring some source.""")

    # Specification configuration
    _allows_refs = False

    _valid_keys = ['variables', 'auth', 'defaults', 'config', 'sources', 'pipelines', 'targets']

    def __init__(self, specification=None, **params):
        self._load_global = params.pop('load_global', True)
        self._yaml_file = specification
        self._root = config.root = os.path.abspath(os.path.dirname(self._yaml_file))
        self._edited = False
        self._debug = params.pop('debug', False)
        super().__init__(**params)

        # Initialize from spec
        config.load_local_modules()
        self._load_specification(from_file=True)
        self._apps = {}

        # Set up background loading
        self._thread_pool = None
        self._rendered = []

        # Initialize high-level settings
        self.auth = Auth.from_spec(state.spec.get('auth', {}))
        self.config = Config.from_spec(state.spec.get('config', {}))
        self.defaults = Defaults.from_spec(state.spec.get('defaults', {}))
        self.variables = Variables.from_spec(state.spec.get('variables', {}))
        self.defaults.apply()

        # Load and populate template
        self._template = self.config.construct_template()
        self._create_modal()
        self._create_main()
        self._create_sidebar()
        self._create_header()
        self._populate_template()

        state._apps[pn.state.curdoc] = self
        pn.state.onload(self._render_dashboard)

    ##################################################################
    # Load specification
    ##################################################################

    def _render_dashboard(self):
        try:
            self._materialize_specification()
            self._render()
            self._main[:] = [self._layout]
        except Exception as e:
            self.param.warning(f'Rendering dashboard raised following error:\n\n {type(e).__name__}: {e}')
            self._main.loading = False
            tb = html.escape(traceback.format_exc())
            alert = pn.pane.HTML(
                f'<b>{type(e).__name__}</b>: {e}</br><pre style="overflow-y: scroll">{tb}</pre>',
                css_classes=['alert', 'alert-danger'], sizing_mode='stretch_width'
            )
            self._main[:] = [alert]
        if isinstance(self._layout, pn.Tabs) and self.config.sync_with_url:
            pn.state.location.sync(self._layout, {'active': 'target'})

    def _load_specification(self, from_file=False):
        kwargs = {}
        if from_file or self._yaml is None:
            with open(self._yaml_file) as f:
                self._yaml = f.read()
        elif self._edited:
            kwargs = {'getenv': False, 'getshell': False, 'getoauth': False}
        state.spec = self.validate(load_yaml(self._yaml, **kwargs))
        state.resolve_views()

    def _load_target(self, target_spec):
        target_spec = dict(target_spec)
        if 'reloadable' not in target_spec:
            target_spec['reloadable'] = self.config.reloadable
        if 'auto_update' not in target_spec:
            target_spec['auto_update'] = self.config.auto_update
        target = Target.from_spec(target_spec, application=self)
        if isinstance(self._layout, pn.Tabs):
            target.show_title = False
        target.start()
        return target

    def _background_load(self, i, spec):
        self.targets[i] = target = self._load_target(spec)
        return target

    def _materialize_specification(self, force=False):
        if force or self._load_global or not state.global_sources:
            state.load_global_sources(clear_cache=force)
        if force or self._load_global or not state.pipelines:
            state.load_pipelines(auto_update=self.config.auto_update)
        if not self.auth.authorized:
            self.targets = []
            return
        targets = []
        target_specs = state.spec.get('targets', [])
        ntargets = len(target_specs)
        if isinstance(self._layout, pn.Tabs) and self.config.background_load:
            self._thread_pool = ThreadPoolExecutor(max_workers=ntargets-1)
        for i, target_spec in enumerate(target_specs):
            if state.loading_msg:
                state.loading_msg.object = (
                    f"Loading target {target_spec['title']!r} ({i + 1}/{ntargets})..."
            )
            if isinstance(self._layout, pn.Tabs) and i != self._layout.active:
                if self.config.background_load:
                    item = self._thread_pool.submit(self._background_load, i, target_spec)
                else:
                    item = None
                self._rendered.append(False)
            else:
                item = self._load_target(target_spec)
                self._rendered.append(True)
            targets.append(item)
        self.targets[:] = targets
        if self._thread_pool is not None:
            self._thread_pool.shutdown()

    ##################################################################
    # Create UI
    ##################################################################

    def _create_header(self):
        self._header = pn.Row()
        self._reload_button = IconButton(
            icon='fa-sync', size=18, margin=0
        )
        self._reload_button.on_click(self._reload)
        menu_items = []
        for yml in config.yamls:
            with open(yml) as f:
                spec = load_yaml(f.read())
            endpoint = os.path.basename(yml).split('.')[0]
            title = spec.get('config', {}).get('title', endpoint)
            i = 1
            while title in self._apps:
                i += 1
                title = f'{title} {i}'
            menu_items.append(title)
            self._apps[title] = endpoint
        if len(config.yamls) > 1:
            self._reload_button.margin = (7, 0, 0, 0)
            self._menu_button = pn.widgets.MenuButton(
                name='Select dashboard', items=menu_items, width=200,
                align='center', margin=(0, 25, 0, 0)
            )
            self._menu_button.on_click(self._navigate)
            self._header.append(self._menu_button)
        if self.config.reloadable:
            self._header.append(self._reload_button)
        if self.config.editable:
            self._header.append(self._edit_button)
        if 'auth' in state.spec:
            logout = IconButton(
                icon='fa-sign-out-alt', size=18, margin=0
            )
            logout.js_on_click(code="""
            window.location.href = '/logout'
            """)
            self._header.extend([
                pn.layout.HSpacer(),
                f'<b><font size="4.5em">User: {pn.state.user}</font></b>',
                logout
            ])

    def _create_main(self):
        layout_kwargs = {'sizing_mode': 'stretch_both', 'margin': 10}
        if self.config.layout is pn.Tabs:
            layout_kwargs['dynamic'] = True
        elif self.config.layout is pn.GridBox:
            layout_kwargs['ncols'] = self.config.ncols
        self._layout = self.config.layout(**layout_kwargs)
        style = {'text-align': 'center', 'font-size': '1.8em', 'font-weight': 'bold'}
        state.loading_msg = pn.pane.HTML(
            'Loading...', align='center', style=style, width=400, height=400
        )
        self._loading = pn.Column(
            pn.layout.HSpacer(), state.loading_msg, pn.layout.HSpacer(),
            sizing_mode='stretch_both'
        )
        self._main = pn.Column(
            self._loading, loading=True, sizing_mode='stretch_both',
        )
        if isinstance(self._layout, pn.Tabs):
            self._layout.param.watch(self._activate_filters, 'active')

    def _create_modal(self):
        self._editor = pn.widgets.Ace(
            value=self._yaml, filename=self._yaml_file,
            sizing_mode='stretch_both', min_height=600,
            theme='monokai'
        )
        self._edit_button = pn.widgets.Button(
            name='✎', width=50, css_classes=['reload'], margin=0,
            align='center'
        )
        self._editor.param.watch(self._edit, 'value')
        self._edit_button.js_on_click(code="""
        var modal = document.getElementById("pn-Modal")
        modal.style.display = "block"
        """)
        self._modal = pn.Column(
            f'## Edit {os.path.basename(self._yaml_file)}',
            self._editor, sizing_mode='stretch_both'
        )

    def _create_sidebar(self):
        self._sidebar = pn.Accordion(margin=0, sizing_mode='stretch_width')

    def _populate_template(self):
        self._template.modal[:] = [self._modal]
        self._template.main[:] = [self._main]
        self._template.header[:] = [self._header]
        self._template.sidebar[:] = [self._sidebar]

    ##################################################################
    # Rendering API
    ##################################################################

    @catch_and_notify
    def _activate_filters(self, event):
        target = self.targets[event.new]
        rendered = self._rendered[event.new]
        if not rendered:
            spec = state.spec['targets'][event.new]
            self._set_loading(spec['title'])
            if isinstance(target, Future):
                target = target.result()
            else:
                target = self._load_target(spec)
            self.targets[event.new] = target
            self._layout[event.new] = target.panels
            self._rendered[event.new] = True
        self._render_filters()
        self._layout.loading = False

    def _edit(self, event):
        self._yaml = event.new
        self._edited = True

    def _navigate(self, event):
        app = self._apps[event.new]
        pn.state.location.pathname = f'/{app}'
        pn.state.location.reload = True

    def _set_loading(self, name=''):
        state.loading_msg.object = f'<b>Reloading {name}...</b>'
        if isinstance(self._layout, pn.GridBox):
            items = [pn.pane.HTML(width=self._layout[i].width)
                     for i in range(self._layout.ncols) if i < len(self._layout)]
            index = int(min(self._layout.ncols, (len(self._layout)-1)) / 2)
            if items:
                items[index] = self._loading
            else:
                items = [self._loading]
        elif isinstance(self._layout, pn.Tabs):
            items = list(zip(self._layout._names, list(self._layout.objects)))
            tab_name = items[self._layout.active][0]
            if name and tab_name != name:
                return
            items[self._layout.active] = (tab_name, self._loading)
        else:
            items = [self._loading]
        self._layout[:] = items
        self._layout.loading = True

    def _open_modal(self, event):
        self._template.open_modal()

    def _get_global_filters(self):
        views = []
        filters = []
        for target in self.targets:
            if target is None or isinstance(target, Future):
                continue
            for pipeline in target._pipelines.values():
                for filt in pipeline.filters:
                    if ((all(isinstance(target, (type(None), Future)) or any(filt in p.filters for p in target._pipelines.values())
                             for target in self.targets) and
                         filt.panel is not None) and filt.shared) and filt not in filters:
                        views.append(filt.panel)
                        filters.append(filt)
        if not self.config.auto_update:
            button = pn.widgets.Button(name='Apply Update')
            def update_pipelines(event):
                for target in self.targets:
                    if target is None or isinstance(target, Future):
                        continue
                    for pipeline in target._pipelines.values():
                        pipeline.param.trigger('update')
            button.on_click(update_pipelines)
            views.append(button)
        if not views or len(self.targets) == 1:
            return None, None
        return filters, pn.Column(*views, name='Filters', sizing_mode='stretch_width')

    def _render_filters(self):
        self._global_filters, global_panel = self._get_global_filters()
        filters = [] if global_panel is None else [global_panel]
        global_refs = [ref.split('.')[1] for ref in state.global_refs]
        self._variable_panel = self.variables.panel(global_refs)
        apply_button = not self.config.auto_update or len(self.targets) == 1
        if self._variable_panel is not None:
            filters.append(self._variable_panel)
        for i, target in enumerate(self.targets):
            if isinstance(self._layout, pn.Tabs) and i != self._layout.active:
                continue
            elif target is None or isinstance(target, Future):
                spec = state.spec['targets'][i]
                panel = pn.Column(name=spec['title'])
            else:
                panel = target.get_filter_panel(skip=self._global_filters, apply_button=apply_button)
            if panel is not None:
                filters.append(panel)
        self._sidebar[:] = filters

    def _render_targets(self):
        items = []
        for target, spec in zip(self.targets, state.spec.get('targets', [])):
            if target is None or isinstance(target, Future):
                panel = pn.Column(name=spec['title'])
            else:
                panel = target.panels
            items.append(panel)
        self._layout[:] = items

    def _render_unauthorized(self):
        auth_keys = list(state.spec.get('auth'))
        if len(auth_keys) == 1:
            auth_keys = repr(auth_keys[0])
        else:
            auth_keys = [repr(k) for k in auth_keys]
        if pn.state.user_info is None:
            error = """
            ## Authorization Error

            Cannot verify permissioning since OAuth is not enabled.
            """
        else:
            error = f"""
            ## Unauthorized User

            {pn.state.user} is not authorized. Ensure {pn.state.user}
            is permissioned correctly on the {auth_keys} field(s) in
            OAuth user data.
            """
        alert = pn.pane.Alert(error, alert_type='danger')
        if isinstance(self._layout, pn.Tabs):
            alert = ('Authorization Denied', alert)
        self._layout[:] = [alert]

    def _render(self):
        if self.auth.authorized:
            self._render_filters()
            if isinstance(self._layout, pn.Tabs):
                count = 1 + bool(self._global_filters) + bool(self._variable_panel)
                active = list(range(count))
            else:
                active = list(range(len(self._sidebar)))
            self._sidebar.active = active
            self._render_targets()
        else:
            self._render_unauthorized()
        self._main.loading = False

    def _reload(self, *events):
        self._load_specification()
        self._materialize_specification(force=True)
        self._render()

    ##################################################################
    # Validation API
    ##################################################################

    @classmethod
    def _validate_pipelines(cls, pipeline_specs, spec, context):
        if 'pipelines' not in context:
            context['pipelines'] = {}
        return cls._validate_dict_subtypes('pipelines', Pipeline, pipeline_specs, spec, context, context['pipelines'])

    @classmethod
    def _validate_sources(cls, source_specs, spec, context):
        if 'sources' not in context:
            context['sources'] = {}
        return cls._validate_dict_subtypes('sources', Source, source_specs, spec, context, context['sources'])

    @classmethod
    def _validate_targets(cls, target_specs, spec, context):
        if 'targets' not in context:
            context['targets'] = []
        return cls._validate_list_subtypes('targets', Target, target_specs, spec, context, context['targets'])

    @classmethod
    def _validate_variables(cls, variable_specs, spec, context):
        if 'variables' not in context:
            context['variables'] = {}
        return cls._validate_dict_subtypes('variables', Variable, variable_specs, spec, context, context['variables'])

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def validate(cls, spec):
        """
        Validates the component specification given the validation context.

        Arguments
        -----------
        spec: dict
          The specification for the component being validated.

        Returns
        --------
        Validated specification.
        """
        cls._validate_keys(spec)
        cls._validate_required(spec)
        return cls._validate_spec(spec)

    def layout(self):
        """
        Returns a layout of the dashboard contents.
        """
        spinner = pn.indicators.LoadingSpinner(width=40, height=40)
        pn.state.sync_busy(spinner)

        title_html = f'<font color="white"><h1>{self.config.title}</h1></font>'
        return pn.Column(
            pn.Row(
                pn.pane.HTML(title_html),
                pn.layout.HSpacer(),
                spinner,
                background='#00aa41'
            ),
            pn.Row(
                pn.Column(self._sidebar, width=300),
                self._layout
            )
        )

    def to_spec(self):
        """
        Exports the full specification to reconstruct this component.

        Returns
        -------
        Declarative specification of this component.
        """
        spec = {}
        spec['variables'] = {
            name: variable.to_spec(context=spec)
            for name, variable in state.variables._vars.items()
        }
        spec['sources'] = {
            name: source.to_spec(context=spec)
            for name, source in state.sources.items()
        }
        spec['pipelines'] = {
            name: pipeline.to_spec(context=spec)
            for name, pipeline in state.pipelines.items()
        }
        spec.update(super().to_spec(context=spec))
        spec['targets'] = [
            tspec if target is None else target
            for target, tspec in zip(spec['targets'], state.spec['targets'])
        ]
        return {k: v for k, v in spec.items() if v != {}}

    def show(self, **kwargs):
        """
        Opens the dashboard in a new window.
        """
        self._template.show(title=self.config.title, **kwargs)

    def servable(self):
        """
        Marks the dashboard for serving with `panel serve`.
        """
        self._template.servable(title=self.config.title)



import datetime as dt
import os

import pandas as pd
import pytest

from lumen.sources.intake_sql import IntakeSQLSource
from lumen.transforms.sql import SQLGroupBy


@pytest.fixture
def source():
    root = os.path.dirname(__file__)
    intake_sql_source = IntakeSQLSource(
        uri=os.path.join(root, 'catalog.yml'), root=root
    )
    return intake_sql_source


@pytest.fixture
def source_tables():
    df_test = pd._testing.makeMixedDataFrame()
    df_test_sql = pd._testing.makeMixedDataFrame()
    df_test_sql_none = pd._testing.makeMixedDataFrame()
    df_test_sql_none['C'] = ['foo1', None, 'foo3', None, 'foo5']
    tables = {
        'test': df_test,
        'test_sql': df_test_sql,
        'test_sql_with_none': df_test_sql_none,
    }
    return tables


def test_intake_sql_resolve_module_type():
    assert IntakeSQLSource._get_type('lumen.sources.intake_sql.IntakeSQLSource') is IntakeSQLSource
    assert IntakeSQLSource.source_type == 'intake_sql'


def test_intake_sql_get_tables(source, source_tables):
    tables = source.get_tables()
    assert tables == list(source_tables.keys())
    for table in tables:
        pd.testing.assert_frame_equal(source.get(table), source_tables[table])


def test_intake_sql_get_schema(source):
    expected_sql = {
        'A': {'inclusiveMaximum': 4.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'B': {'inclusiveMaximum': 1.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'C': {'enum': ['foo1', 'foo2', 'foo3', 'foo4', 'foo5'], 'type': 'string'},
        'D': {
            'format': 'datetime',
            'inclusiveMaximum': '2009-01-07 00:00:00',
            'inclusiveMinimum': '2009-01-01 00:00:00',
            'type': 'string'
        }
    }
    expected_csv = dict(expected_sql, D={
        'format': 'datetime',
        'inclusiveMaximum': '2009-01-07T00:00:00',
        'inclusiveMinimum': '2009-01-01T00:00:00',
        'type': 'string'
    })
    assert source.get_schema('test_sql') == expected_sql
    assert list(source._schema_cache.keys()) == ['test_sql']
    assert source.get_schema('test') == expected_csv
    assert list(source._schema_cache.keys()) == ['test_sql', 'test']


def test_intake_sql_get_schema_with_none(source):
    expected_sql = {
        'A': {'inclusiveMaximum': 4.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'B': {'inclusiveMaximum': 1.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'C': {'enum': ['foo1', None, 'foo3', 'foo5'], 'type': 'string'},
        'D': {
            'format': 'datetime',
            'inclusiveMaximum': '2009-01-07 00:00:00',
            'inclusiveMinimum': '2009-01-01 00:00:00',
            'type': 'string'
        }
    }
    assert source.get_schema('test_sql_with_none') == expected_sql
    assert list(source._schema_cache.keys()) == ['test_sql_with_none']


def test_intake_sql_get_schema_cache(source):
    source.get_schema('test_sql')
    assert 'test_sql' in source._schema_cache


@pytest.mark.parametrize(
    "table_column_value_type", [
        ('test_sql', 'A', 1, 'single_value'),
        ('test_sql', 'A', (1, 3), 'range'),
        ('test_sql', 'A', [(0, 1), (3, 4)], 'range_list'),
        ('test_sql', 'C', 'foo2', 'single_value'),
        ('test_sql', 'C', ['foo1', 'foo3'], 'list'),
        ('test_sql', 'D', dt.datetime(2009, 1, 2), 'single_value'),
        ('test_sql', 'D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('test_sql', 'D', dt.date(2009, 1, 2), 'date'),
        ('test_sql', 'D', (dt.date(2009, 1, 2), dt.date(2009, 1, 5)), 'date_range'),
        ('test_sql_with_none', 'C', None, 'single_value'),
        ('test_sql_with_none', 'C', [None, 'foo5'], 'list'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_intake_sql_filter(source, table_column_value_type, dask, expected_filtered_df):
    table, column, value, _ = table_column_value_type
    kwargs = {column: value}
    filtered = source.get(table, __dask=dask, **kwargs)
    pd.testing.assert_frame_equal(filtered, expected_filtered_df.reset_index(drop=True))


def test_intake_sql_transforms(source, source_tables):
    df_test_sql = source_tables['test_sql']
    transforms = [SQLGroupBy(by=['B'], aggregates={'SUM': 'A'})]
    transformed = source.get('test_sql', sql_transforms=transforms)
    expected = df_test_sql.groupby('B')['A'].sum().reset_index()
    pd.testing.assert_frame_equal(transformed, expected)


def test_intake_sql_transforms_cache(source, source_tables):
    df_test_sql = source_tables['test_sql']
    transforms = [SQLGroupBy(by=['B'], aggregates={'SUM': 'A'})]

    source.get('test_sql', sql_transforms=transforms)
    cache_key = source._get_key('test_sql', sql_transforms=transforms)
    assert cache_key in source._cache

    expected = df_test_sql.groupby('B')['A'].sum().reset_index()
    pd.testing.assert_frame_equal(source._cache[cache_key], expected)

    cache_key = source._get_key('test_sql', sql_transforms=transforms)
    assert cache_key in source._cache


def test_intake_sql_clear_cache(source):
    source.get('test_sql')
    source.get_schema('test_sql')
    assert len(source._cache) == 1
    assert len(source._schema_cache) == 1
    source.clear_cache()
    assert len(source._cache) == 0
    assert len(source._schema_cache) == 0



import importlib
import os
import sys
import weakref

import panel as pn
import param as param

from panel.template import DarkTheme, DefaultTheme
from panel.template.base import BasicTemplate
from panel.widgets.indicators import Indicator

from .validation import ValidationError, match_suggestion_message


class ConfigDict(dict):
    def __init__(self, name, **kwargs):
        self.name = name
        super().__init__(**kwargs)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as e:
            msg = f"{self.name} with name '{key}' was not found."
            msg = match_suggestion_message(key, list(self), msg)
            raise ValidationError(msg) from e


_INDICATORS = {k.lower(): v for k, v in param.concrete_descendents(Indicator).items()}
_INDICATORS = ConfigDict("Indicator", **_INDICATORS)

_LAYOUTS = {
    'accordion': pn.Accordion,
    'column'   : pn.Column,
    'grid'     : pn.GridBox,
    'row'      : pn.Row,
    'tabs'     : pn.Tabs
}

try:
    _DEFAULT_LAYOUT = _LAYOUTS['flex'] = pn.FlexBox
except Exception:
    _DEFAULT_LAYOUT = pn.GridBox

_LAYOUTS = ConfigDict("Layout", **_LAYOUTS)


_TEMPLATES = {
    k[:-8].lower(): v for k, v in param.concrete_descendents(BasicTemplate).items()
}
_TEMPLATES = ConfigDict("Template", **_TEMPLATES)

_THEMES = {'default': DefaultTheme, 'dark': DarkTheme}
_THEMES = ConfigDict("Theme", **_THEMES)


class _config(param.Parameterized):
    """
    Stores shared configuration for the entire Lumen application.
    """

    _root = param.String(default=None, doc="""
      Root directory relative to which data and modules are loaded.""")

    _dev = param.Boolean(default=None, doc="""
      Whether Lumen should run in development mode.""")

    template_vars = param.Dict(default={}, doc="""
      Template variables which may be referenced in a dashboard yaml
      specification.""")

    yamls = param.List(default=[], doc="""
      List of yaml files currently being served.""")

    _modules = {}

    _session_config = weakref.WeakKeyDictionary()

    @property
    def root(self):
        if pn.state.curdoc and pn.state.curdoc in self._session_config:
            session_config = self._session_config[pn.state.curdoc]
            if 'root' in session_config:
                return session_config['root']
        if self._root:
            return self._root
        return os.getcwd()

    @root.setter
    def root(self, root):
        if pn.state.curdoc:
            if pn.state.curdoc in self._session_config:
                session_config = self._session_config[pn.state.curdoc]
            else:
                self._session_config[pn.state.curdoc] = session_config = {}
            session_config['root'] = root
        else:
            self._root = root

    def load_local_modules(self):
        """
        Loads local modules containing custom components and templates.
        """
        for imp in ('filters', 'sources', 'transforms', 'template', 'views'):
            path = os.path.join(self.root, imp+'.py')
            if not os.path.isfile(path):
                continue
            spec = importlib.util.spec_from_file_location(f"local_lumen.{imp}", path)
            if path in self._modules:
                continue
            self._modules[path] = module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        templates = param.concrete_descendents(BasicTemplate)
        _TEMPLATES.update({
            k[:-8].lower(): v for k, v in templates.items()
        })

    @property
    def dev(self):
        """
        Whether the application was launched in development mode.
        """
        if self._dev is not None:
            return self._dev
        return '--dev' in sys.argv or '--autoreload' in sys.argv

    @dev.setter
    def dev(self, dev):
        self._dev = dev


config = _config()



import datetime as dt

import numpy as np
import pandas as pd
import param
import requests

from ae5_tools.api import AEAdminSession, AEUserSession
from panel import state

from ..util import get_dataframe_schema
from .base import Source, cached, cached_schema


class AE5Source(Source):
    """
    The AE5Source provides a number of tables, which allow monitoring
    the nodes, deployments, sessions and resource profiles of an
    Anaconda Enterprise 5 installation.
    """

    hostname = param.String(doc="URL of the AE5 host.")

    username = param.String(doc="Username to authenticate with AE5.")

    password = param.String(doc="Password to authenticate with AE5.")

    admin_username = param.String(doc="Username to authenticate admin with AE5.")

    admin_password = param.String(doc="Password to authenticate admin with AE5.")

    k8s_endpoint = param.String(default='k8s')

    pool_size = param.Integer(default=100, doc="""
      Size of HTTP socket pool.""")

    private = param.Boolean(default=True, doc="""
      Whether to limit the deployments visible to a user based on
      their authorization.""")

    source_type = 'ae5'

    _deployment_columns = [
        'id', 'name', 'url', 'owner', 'resource_profile', 'public', 'state',
        'cpu', 'cpu_percent', 'memory', 'memory_percent', 'uptime', 'restarts',
        'node'
    ]

    _job_columns = [
        'id', 'name', 'owner', 'command', 'revision', 'resource_profile',
        'created', 'updated', 'state', 'project_id', 'project_name',
        'goal_state', 'status_text', 'url', 'schedule', 'source',
    ]

    _session_columns = [
        'id', 'name', 'url', 'owner', 'resource_profile', 'state', 'node',
    ]

    _allocation_columns = [
        'id', 'name', 'url', 'owner', 'resource_profile', 'state', 'node',
        'type', 'node/mem', 'node/cpu', 'node/gpu', 'resource/cpu',
        'resource/mem', 'resource/gpu', 'allocation/mem_pct',
        'allocation/cpu_pct', 'allocation/gpu_pct'
    ]

    _tables = [
        'deployments', 'nodes', 'resources', 'sessions', 'jobs',
        'resource_allocations'
    ]

    _units = {
        'm': 0.001,
        None: 1,
        'Ki': 1024,
        'Mi': 1024**2,
        'Gi': 1024**3,
        'Ti': 1024**4
    }

    def __init__(self, **params):
        super().__init__(**params)
        self._session = AEUserSession(
            self.hostname, self.username, self.password, persist=False,
            k8s_endpoint=self.k8s_endpoint
        )
        if self.admin_username and self._user:
            self._admin_session = AEAdminSession(
                self.hostname, self.admin_username, self.admin_password
            )
            self._admin_session.authorize()
            try:
                # Try making more optimized query
                user_info = self._admin_session.user_info(self._user, include_login=False)
            except Exception:
                user_info = self._admin_session.user_info(self._user)
            user_id = user_info['id']
            roles = self._admin_session._get(f'users/{user_id}/role-mappings/realm/composite')
            self._is_admin = any(role['name'] == 'ae-admin' for role in roles)
            self._groups = [g['name'] for g in self._admin_session._get(f'users/{user_id}/groups')]
        else:
            self._is_admin = False
            self._groups = []
        adapter = requests.adapters.HTTPAdapter(pool_maxsize=self.pool_size)
        self._session.session.mount('https://', adapter)
        self._session.session.mount('http://', adapter)

    @classmethod
    def _convert_value(cls, value):
        if value == '0':
            return 0
        val_unit = [m for m in cls._units if m is not None and value.endswith(m)]
        val_unit = val_unit[0] if val_unit else None
        if not val_unit:
            return float(value)
        scale_factor = cls._units[val_unit]
        return float(value[:-len(val_unit)]) * scale_factor

    @property
    def _user(self):
        return state.headers.get('Anaconda-User') if self.private else None

    def _process_deployment(self, deployment):
        container_info = deployment.get('_k8s', {}).get('containers', {})
        if 'app' not in container_info:
            nan = float('NaN')
            deployment['cpu'] = nan
            deployment['memory'] = nan
            deployment['restarts'] = nan
            deployment['uptime'] = '-'
            return deployment

        container = container_info['app']
        limits = container['limits']
        usage = container['usage']

        # CPU Usage
        cpu_cap = self._convert_value(limits['cpu'])
        if 'cpu' in usage:
            cpu = self._convert_value(usage['cpu'])
            deployment['cpu'] = cpu
            deployment['cpu_percent'] = round((cpu / cpu_cap)*100, 2)
        else:
            deployment['cpu_percent'] = float('nan')
            deployment['cpu'] = float('nan')

        # Memory usage
        mem_cap = self._convert_value(limits['memory'])
        if 'memory' in usage:
            mem = self._convert_value(usage['memory'])
            deployment['memory'] = mem
            deployment['memory_percent'] = round((mem / mem_cap)*100, 2)
        else:
            deployment['memory'] = float('nan')
            deployment['memory_percent'] = float('nan')

        # Uptime
        started = container.get('since')
        deployment["restarts"] = container['restarts']
        if started:
            started_dt = dt.datetime.fromisoformat(started.replace('Z', ''))
            uptime = str(dt.datetime.now()-started_dt)
            deployment["uptime"] = uptime[:uptime.index('.')]
        else:
            deployment["uptime"] = '-'

        return deployment

    def _process_nodes(self, node):
        caps = {
            'cpu': self._convert_value(node['capacity/cpu']),
            'gpu': self._convert_value(node['capacity/gpu']),
            'mem': self._convert_value(node['capacity/mem']),
            'pod': int(node['capacity/pod'])
        }
        for column in node.index:
            if 'capacity' in column or '/' not in column:
                continue
            vtype = column.split('/')[1]
            cap = caps[vtype]
            value = node[column]
            if vtype != 'pod':
                value = self._convert_value(value)
            node[column] = value
            node[f'{column}_percent'] = 0 if cap == 0 else round((value/caps[vtype])*100, 2)
        return node

    def _get_deployments(self):
        user = self._user
        deployments = self._session.deployment_list(
            k8s=True, format='dataframe', collaborators=bool(user)
        ).apply(self._process_deployment, axis=1)
        if user is None or self._is_admin:
            return deployments[self._deployment_columns]
        return deployments[
            deployments.public |
            (deployments.owner == user) |
            deployments._collaborators.apply(
                lambda cs: any(c['id'] in self._groups if c['type'] == 'group' else c['id'] == user
                               for c in cs)
            )
        ][self._deployment_columns]

    def _get_nodes(self):
        nodes = self._session.node_list(format='dataframe').apply(self._process_nodes, axis=1)
        return nodes[[c for c in nodes.columns if not c.startswith('_')]]

    def _get_resources(self):
        return self._session.resource_profile_list(format='dataframe')

    def _get_sessions(self):
        sessions = self._session.session_list(format='dataframe', k8s=True)
        if self._user and not self._is_admin:
            sessions = sessions[sessions.owner==self._user]
        return sessions[self._session_columns]

    def _get_jobs(self):
        jobs = self._session.job_list(format='dataframe')
        if self._user and not self._is_admin:
            jobs = jobs[jobs.owner==self._user]
        return jobs[self._job_columns]

    def _get_resource_allocations(self):
        """
        Combines deployment resource profiles with resource information
        and node information to allow computing total and per node
        resource allocations.
        """
        resources = self.get('resources').copy()  # Resource profiles
        nodes = self.get('nodes').copy()  # Cluster nodes
        jobs = self.get('jobs').copy()
        deployments = self.get('deployments').copy()
        sessions = self.get('sessions').copy()

        # Pre-process the nodes, resources and jobs tables
        nodes['capacity/gpu'] = nodes['capacity/gpu'].astype(float)
        nodes['capacity/cpu'] = nodes['capacity/cpu'].astype(float)
        nodes['capacity/mem'] = nodes['capacity/mem'].apply(self._convert_value)
        nodes_cols = ['capacity/mem', 'capacity/cpu', 'capacity/gpu', 'name']
        nodes = nodes[nodes_cols]
        nodes = nodes.rename(columns={
            'name': 'node',
            'capacity/mem': 'node/mem',
            'capacity/cpu': 'node/cpu',
            'capacity/gpu': 'node/gpu',
        })
        resources['cpu'] = resources['cpu'].astype(float)
        resources['gpu'] = resources['gpu'].astype(float)
        resources['memory'] = resources['memory'].apply(self._convert_value)
        resources = resources.set_index('name')
        resources_cols = ['cpu', 'memory', 'gpu']
        resources = resources[resources_cols]
        resources = resources.rename(columns={'memory': 'mem'})
        resources = resources.add_prefix('resource/')
        # ae5tools doesn't yet return a note column for the jobs
        jobs['node'] = np.nan

        deployments['type'] = 'deployment'
        jobs['type'] = 'job'
        sessions['type'] = 'session'

        # Concat the deployments, sessions and jobs in a single utilizations table
        common_cols = ['id', 'name', 'url', 'owner', 'resource_profile', 'state', 'node', 'type']
        utilizations = pd.concat([deployments[common_cols], sessions[common_cols], jobs[common_cols]])
        utilizations = pd.merge(utilizations, nodes, left_on='node', right_on='node', how='outer')
        # To remove the orchestring node that has no attached jobs, resources or deployments.
        utilizations = utilizations[~utilizations['id'].isna()]

        allocations = pd.merge(utilizations, resources, left_on='resource_profile', right_index=True, how='left')

        allocations['allocation/mem_pct'] = 100 * allocations['resource/mem'] / allocations['node/mem']
        allocations['allocation/cpu_pct'] = 100 * allocations['resource/cpu'] / allocations['node/cpu']
        allocations['allocation/gpu_pct'] = 100 * allocations['resource/gpu'] / allocations['node/gpu']

        return allocations[self._allocation_columns]

    @cached(with_query=False)
    def get(self, table, **query):
        if table not in self._tables:
            raise ValueError(f"AE5Source has no '{table}' table, choose from {repr(self._tables)}.")
        return getattr(self, f'_get_{table}')()

    @cached_schema
    def get_schema(self, table=None):
        schemas = {}
        for t in self._tables:
            if table is None or t == table:
                schemas[t] = get_dataframe_schema(self.get(t))['items']['properties']
        return schemas if table is None else schemas[table]

    def get_tables(self):
        return self._tables



import param as _param

from .config import config  # noqa
from .dashboard import Dashboard  # noqa
from .filters import Filter  # noqa
from .sources import Source  # noqa
from .transforms import Transform  # noqa
from .views import View  # noqa

__version__ = str(_param.version.Version(
    fpath=__file__, archive_commit="$Format:%h$", reponame="lumen"))



import pytest

from lumen.validation import ValidationError
from lumen.views import Download, View


@pytest.mark.parametrize(
    "spec,msg",
    (
        (
            "csv",
            None,
        ),
        (
            {"format": "csv"},
            None,
        ),
        (
            {"formats": "csv"},
            "The Download component requires 'format' parameter to be defined",
        ),
        (
            {"format": "csvs"},
            "Download component 'format' value failed validation: csvs",
        ),
    ),
    ids=["correct1", "correct2", "missing_required", "wrong_format"],
)
def test_target_Download(spec, msg):
    if msg is None:
        if isinstance(spec, str):
            assert Download.validate(spec) == {"format": spec}
        else:
            assert Download.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Download.validate(spec)


@pytest.mark.parametrize(
    "spec,msg",
    (
        (
            {"type": "download", "format": "csv"},
            None,
        ),
        (
            {"type": "download", "formats": "csv"},
            "The DownloadView component requires 'format' parameter to be defined",
        ),
        (
            {"type": "download", "format": "csvs"},
            "DownloadView component 'format' value failed validation: csvs",
        ),
        (
            {"format": "csv"},
            "View component specification did not declare a type",
        ),
        (
            {"type": "downloads", "format": "csv"},
            "View component specification declared unknown type 'downloads'",
        ),
    ),
    ids=["correct", "missing_required", "wrong_format", "missing_type", "wrong_type"],
)
def test_target_View(spec, msg):
    if msg is None:
        assert View.validate(spec.copy()) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            View.validate(spec)



import datetime as dt
import warnings

from functools import partial
from io import BytesIO, StringIO
from itertools import product

import panel as pn
import param

from panel.viewable import Layoutable, Viewer

from .base import Component
from .config import _LAYOUTS
from .filters import FacetFilter, Filter
from .panel import IconButton
from .pipeline import Pipeline
from .sources import Source
from .state import state
from .util import catch_and_notify, extract_refs
from .validation import ValidationError, match_suggestion_message
from .views import DOWNLOAD_FORMATS, View


class Card(Viewer):
    """
    A Card renders a layout of multiple views given a layout specification.
    """

    layout = param.ClassSelector(default='column', class_=(str, list, dict), doc="""
        Defines the layout of the views in the monitor target. Can be
        'column', 'row', 'grid', 'row' or a nested list of indexes
        corresponding to the views, e.g. [[0, 1], [2]] will create
        a Column of one row containing views 0 and 1 and a second Row
        containing view 2.""" )

    title = param.String(doc="A title for this Card.")

    views = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of View objects.""")

    def __init__(self, **params):
        self.kwargs = params.pop('kwargs', {})
        super().__init__(**params)
        params = {k: v for k, v in self.kwargs.items() if k in pn.Card.param}
        self._card = pn.Card(
            title=self.title, name=self.title, collapsible=False, **params
        )
        self._card[:] = [self._construct_layout()]

    def __panel__(self):
        return self._card

    def _construct_layout(self):
        layout = self.layout
        if isinstance(layout, list):
            view_size = len(self.views)
            item = pn.Column(sizing_mode='stretch_both')
            for row_spec in layout:
                row = pn.Row(sizing_mode='stretch_width')
                for index in row_spec:
                    if isinstance(index, int):
                        if index >= view_size:
                            raise ValueError(
                                f"Layout specification for '{self.title}' target references "
                                f"out-of-bounds index ({index}) even though the maximum "
                                f"available index is {view_size - 1}."
                            )
                        view = self.views[index]
                    else:
                        matches = [view for view in self.views if view.name == index]
                        if matches:
                            view = matches[0]
                        else:
                            raise ValueError(
                                f"Target could not find named view '{index}'."
                            )
                    row.append(view.panel)
                item.append(row)
        else:
            kwargs = dict(self.kwargs)
            if isinstance(layout, dict):
                layout_kwargs = dict(layout)
                layout = layout_kwargs.pop('type')
                kwargs.update(layout_kwargs)
            if layout == 'grid' and 'ncols' not in kwargs:
                kwargs['ncols'] = 2
            layout_type = _LAYOUTS[layout]
            layout_params = list(layout_type.param.params())
            kwargs = {k: v for k, v in kwargs.items() if k in layout_params}
            item = layout_type(*(view.panel for view in self.views), **kwargs)
        return item

    @classmethod
    def from_spec(cls, spec, filters=None, pipelines={}):
        spec = dict(spec)
        view_specs = spec.get('views', [])
        spec['views'] = views = []
        for view in view_specs:
            if isinstance(view_specs, dict):
                view_spec = dict(view_specs[view], name=view)
            else:
                view_spec = view
            if 'table' in view_spec and view_spec['table'] in pipelines:
                pipeline = pipelines[view_spec['table']]
            elif len(pipelines) == 1:
                pipeline = list(pipelines.values())[0]
                if 'pipeline' in view_spec:
                    del view_spec['pipeline']
            if filters:
                pipeline = pipeline.chain(filters=list(filters))
            view = View.from_spec(view_spec, pipeline=pipeline)
            views.append(view)
        if filters:
            spec['title'] = ' '.join([f'{f.label}: {f.value}' for f in filters])
        return cls(**spec)

    def rerender(self):
        self._card[:] = [self._construct_layout()]


class Facet(Component):

    by = param.List(default=[], class_=FacetFilter, doc="""
        Fields to facet by.""")

    layout = param.ClassSelector(default=None, class_=(str, dict), doc="""
        How to lay out the facets.""")

    reverse = param.Boolean(default=False, doc="""
        Whether to reverse the sort order.""")

    sort = param.ListSelector(default=[], objects=[], doc="""
        List of fields to sort by.""")

    _valid_keys = 'params'
    _required_keys = ['by']

    def __init__(self, **params):
        super().__init__(**params)
        self._sort_widget = pn.widgets.MultiSelect(
            options=self.param.sort.objects,
            sizing_mode='stretch_width',
            size=len(self.sort),
            value=self.sort,
        )
        self._sort_widget.link(self, value='sort')
        self._reverse_widget = pn.widgets.Checkbox(
            value=self.reverse, name='Reverse', margin=(5, 0, 0, 10)
        )
        self._reverse_widget.link(self, value='reverse')

    def get_sort_key(self, views):
        sort_key = []
        for field in self.sort:
            values = [v.get_value(field) for v in views]
            if values:
                sort_key.append(values[0])
        return tuple(sort_key)

    @param.depends('sort:objects', watch=True)
    def _update_options(self):
        self._sort_widget.options = self.param.sort.objects

    @classmethod
    def from_spec(cls, spec, schema):
        """
        Creates a Facet object from a schema and a set of fields.
        """
        by = []
        for by_spec in spec.pop('by', []):
            if isinstance(by_spec, str):
                f = Filter.from_spec({'type': 'facet', 'field': by_spec}, schema)
            elif isinstance(by_spec, dict):
                f = Filter.from_spec(dict(by_spec, type='facet'), schema)
            elif isinstance(by_spec, Filter):
                f = by_spec
            by.append(f)
        sort = spec.pop('sort', [b.field for b in by])
        sorter = cls(by=by, **spec)
        sorter.param.sort.objects = sort
        return sorter

    @property
    def filters(self):
        return product(*[filt.filters for filt in self.by])


class Download(Component, Viewer):
    """
    The Download object controls options to download the Source tables
    in a variety of formats via the Dashboard UI.
    """

    labels = param.Dict(default={}, doc="""
        A dictionary to override the Button label for each table.""")

    format = param.ObjectSelector(default=None, objects=DOWNLOAD_FORMATS, doc="""
        The format to download the data in.""")

    kwargs = param.Dict(default={}, doc="""
        Keyword arguments passed to the serialization function, e.g.
        data.to_csv(file_obj, **kwargs).""")

    pipelines = param.Dict(default={}, doc="""
        Dictionary of pipelines indexed by table name.""")

    tables = param.List(default=[], doc="""
        The list of tables to allow downloading.""")

    _internal_params = ['name', 'pipelines']
    _required_keys = ['format']

    def __init__(self, **params):
        super().__init__(**params)
        default_table = self.tables[0] if self.tables else None
        button_params = {'filename': f'{default_table}.{self.format}'}
        if default_table in self.labels:
            button_params['label'] = self.labels[default_table]
        self._select_download = pn.widgets.Select(
            name='Select table to download', options=self.tables, value=default_table,
            sizing_mode='stretch_width'
        )
        self._select_download.param.watch(self._update_button, 'value')
        self._download_button = pn.widgets.FileDownload(
            callback=self._table_data, align='end',
            sizing_mode='stretch_width', **button_params
        )
        self._layout = pn.Column(
            pn.pane.Markdown('### Download tables', margin=(0, 5, -10, 5)),
            self._download_button,
            pn.layout.Divider(),
            sizing_mode='stretch_width'
        )
        if len(self.tables) > 1:
            self._layout.insert(1, self._select_download)

    @classmethod
    def validate(cls, spec, context=None):
        if isinstance(spec, str):
            spec = {'format': spec}
        return super().validate(spec, context)

    def __panel__(self):
        return self._layout

    def _update_button(self, event):
        params = {'filename': f'{event.new}.{self.format}'}
        if event.new in self.labels:
            params['label'] = self.labels[event.new]
        self._download_button.param.set_param(**params)

    def __bool__(self):
        return self.format is not None

    @catch_and_notify("Download failed")
    def _table_data(self):
        if self.format in ('json', 'csv'):
            io = StringIO()
        else:
            io = BytesIO()
        table = self._select_download.value
        data = self.pipelines[table].data
        if self.format == 'csv':
            data.to_csv(io, **self.kwargs)
        elif self.format == 'json':
            data.to_json(io, **self.kwargs)
        elif self.format == 'xlsx':
            data.to_excel(io, **self.kwargs)
        elif self.format == 'parquet':
            data.to_parquet(io, **self.kwargs)
        io.seek(0)
        return io

    @classmethod
    def from_spec(cls, spec, pipelines):
        """
        Creates a Download object from a specification.

        Parameters
        ----------
        spec : dict
            Specification declared as a dictionary of parameter values.
        pipelines: Dict[str, Pipeline]
            A dictionary of Pipeline objects indexed by table name.
        """
        return cls(pipelines=pipelines, **spec)


class Target(Component):
    """
    A Target renders the results of a Source query using the defined
    set of filters and views.
    """

    auto_update = param.Boolean(default=True, constant=True, doc="""
        Whether changes in filters, transforms and references automatically
        trigger updates in the data or whether an update has to be triggered
        manually using the update event or the update button in the UI."""
    )

    download = param.ClassSelector(class_=Download, default=Download(), doc="""
        The download objects determines whether and how the source tables
        can be downloaded.""")

    facet = param.ClassSelector(class_=Facet, doc="""
        The facet object determines whether and how to facet the cards
        on the target.""")

    layout = param.ClassSelector(default='column', class_=(str, list, dict), doc="""
        Defines the layout of the views in the monitor target. Can be
        'column', 'row', 'grid', 'row' or a nested list of indexes
        corresponding to the views, e.g. [[0, 1], [2]] will create
        a Column of one row containing views 0 and 1 and a second Row
        containing view 2.""" )

    reloadable = param.Boolean(default=True, doc="""
        Whether to allow reloading data target's source using a button.""")

    show_title = param.Boolean(default=True, doc="""
        Whether to show the title in Card headers.""")

    title = param.String(doc="A title for this Target.")

    refresh_rate = param.Integer(default=None, doc="""
        How frequently to refresh the monitor by querying the adaptor.""")

    source = param.ClassSelector(class_=Source, doc="""
        The Source queries the data from some data source.""")

    tsformat = param.String(default="%m/%d/%Y %H:%M:%S")

    views = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of View specifications.""")

    _header_format = '<div style="font-size: 1.5em; font-weight: bold;">{header}</div>'

    _required_keys = ['title', 'views']

    _valid_keys = [
        'config', 'facet_layout', 'sort', # Deprecated
        'layout', 'refresh_rate', 'reloadable', 'show_title', 'title', 'tsformat', # Simple
        'views', 'source', 'filters', 'pipeline', 'facet', 'download' # Objects
    ] + list(Layoutable.param)

    def __init__(self, **params):
        if 'facet' not in params:
            params['facet'] = Facet()
        self._application = params.pop('application', None)
        self._cache = {}
        self._cb = None
        self._scheduled = False
        self._updates = []
        self._update_button = pn.widgets.Button(name='Apply update')
        self._update_button.on_click(self._manual_update)
        self._timestamp = pn.pane.HTML(
            align='center', margin=(10, 0), sizing_mode='stretch_width'
        )
        self._pipelines = params.pop('pipelines', {})
        self.kwargs = {k: v for k, v in params.items() if k not in self.param}
        super().__init__(**{k: v for k, v in params.items() if k in self.param})

        # Render content
        if self.views:
            self._construct_cards()
        self._cards = self.get_cards()

        # Set up watchers
        self.facet.param.watch(self._resort, ['sort', 'reverse'])
        for view in list(self._cache.values())[0].views:
            view.param.watch(self._schedule_rerender, 'rerender')

    def _resort(self, *events):
        self._rerender(update_views=False)

    def _manual_update(self, *events):
        for p in self._pipelines.values():
            p.param.trigger('update')

    def _sync_component(self, component, *events):
        component.param.set_param(**{event.name: event.new for event in events})

    def _construct_cards(self):
        controls, all_transforms = [], []
        linked_views = None
        for facet_filters in self.facet.filters:
            key = tuple(str(f.value) for f in facet_filters)
            self._cache[key] = card = Card.from_spec({
                'kwargs': self.kwargs,
                'layout': self.layout,
                'title': self.title,
                'views': self.views,
            }, filters=facet_filters, pipelines=self._pipelines)
            if linked_views is None:
                for view in card.views:
                    if view.controls:
                        controls.append(view.control_panel)
                    transforms = (
                        view.pipeline.traverse('transforms') +
                        view.pipeline.traverse('sql_transforms')
                    )
                    for transform in transforms:
                        if transform not in all_transforms and transform.controls:
                            controls.append(transform.control_panel)
                            all_transforms.append(transform)
                linked_views = card.views
                continue

            # Only the controls for the first facet is shown so link
            # the other facets to the controls of the first
            for v1, v2 in zip(linked_views, card.views):
                v1.param.watch(partial(self._sync_component, v2), v1.refs)
                for t1, t2 in zip(v1.pipeline.transforms, v2.pipeline.transforms):
                    t1.param.watch(partial(self._sync_component, t2), t1.refs)
                for t1, t2 in zip(v1.pipeline.sql_transforms, v2.pipeline.sql_transforms):
                    t1.param.watch(partial(self._sync_component, t2), t1.refs)
        self._view_controls = pn.Column(*controls, sizing_mode='stretch_width')

    ##################################################################
    # Create UI
    ##################################################################

    def get_cards(self):
        cards = []
        for card in self._cache.values():
            sort_key = self.facet.get_sort_key(card.views)
            if any(view for view in card.views):
                cards.append((sort_key, card))
        if self.facet.sort:
            cards = sorted(cards, key=lambda x: x[0])
            if self.facet.reverse:
                cards = cards[::-1]
        return [card for _, card in cards]

    def get_filter_panel(self, skip=None, apply_button=True):
        skip = list(skip or [])
        views = []

        # Variable controls
        global_refs = state.global_refs
        target_refs = [ref.split('.')[1] for ref in self.refs if ref not in global_refs]
        var_panel = state.variables.panel(target_refs)
        if var_panel is not None:
            views.append(var_panel)

        # Source controls
        source_panel = self.source.panel
        if source_panel:
            views.extend([
                self._header_format.format(header='Source'),
                source_panel,
                pn.layout.Divider()
            ])

        # Download controls
        if self.download:
            views.append(self.download)

        # Filter controls
        filters = []
        for pipeline in self._pipelines.values():
            subpipeline = pipeline
            while subpipeline is not None:
                subfilters = []
                for filt in subpipeline.filters:
                    fpanel = filt.panel
                    if filt not in skip and fpanel is not None:
                        subfilters.append(fpanel)
                        skip.append(filt)
                if subfilters:
                    filters.append(subfilters)
                subpipeline = subpipeline.pipeline
        if filters:
            views.append(self._header_format.format(header='Filters'))
            views.extend([filt for sfilters in filters[::-1] for filt in sfilters])

        # Facet controls
        if self.facet.param.sort.objects:
            if views:
                views.append(pn.layout.Divider(margin=0, height=5))
            views.extend([
                self._header_format.format(header='Sort'),
                self.facet._sort_widget,
                self.facet._reverse_widget
            ])

        # View controls
        if self._view_controls:
            views.append(self._view_controls)

        # Add update button
        if not self.auto_update and apply_button:
            views.append(self._update_button)

        # Reload buttons
        if self.reloadable:
            if views:
                views.append(pn.layout.Divider(margin=0, height=5))
            self._reload_button = IconButton(
                icon='fa-sync', size=18, margin=10
            )
            self._reload_button.on_click(self.update)
            self._timestamp.object = f'Last updated: {dt.datetime.now().strftime(self.tsformat)}'
            reload_panel = pn.Row(
                self._reload_button, self._timestamp, sizing_mode='stretch_width',
                margin=(10, 10, 0, 5)
            )
            views.append(reload_panel)
        return pn.Column(*views, name=self.title, sizing_mode='stretch_width')

    ##################################################################
    # Rendering API
    ##################################################################

    def _schedule_rerender(self, *events):
        for e in events:
            if isinstance(e.obj, View) and e.obj not in self._updates:
                self._updates.append(e.obj)
        if self._scheduled:
            return
        self._scheduled = True
        if pn.state.curdoc:
            pn.state.curdoc.add_next_tick_callback(self._rerender)
        else:
            self._rerender()

    def _rerender_cards(self, cards):
        if self._application:
            self._application._set_loading(self.title)
        for card in cards:
            if any(view in self._updates for view in card.views):
                card.rerender()
        if self._application:
            self._application._render_targets()
        self._updates = []

    def _rerender(self, update_views=True):
        self._scheduled = False
        cards = self.get_cards()
        if update_views:
            self._rerender_cards(cards)
        if cards != self._cards:
            self._cards[:] = cards
            if self._application:
                self._application._render()
        if self._application:
            self._application._layout.loading = False

    ##################################################################
    # Validation API
    ##################################################################

    @classmethod
    def _validate_config(cls, config, spec, context):
        msg = (
            "Passing 'config' to a Target is deprecated use the 'facet' key "
            "on the target instead."
        )
        return cls._deprecation(msg, 'facet', spec, config)

    @classmethod
    def _validate_sort(cls, sort, spec, context):
        msg = (
            "Passing 'sort' to a Target is deprecated use the 'facet' key "
            "on the target instead."
        )
        return cls._deprecation(msg, 'facet', spec, {'sort': sort})

    @classmethod
    def _validate_facet_layout(cls, facet_layout, spec, context):
        msg = (
            "Passing 'facet_layout' to a Target is deprecated use the 'facet' key "
            "on the target instead."
        )
        return cls._deprecation(msg, 'facet', spec, {'facet_layout': facet_layout})

    @classmethod
    def _validate_pipeline(cls, *args, **kwargs):
        return cls._validate_str_or_spec('pipeline', Pipeline, *args, **kwargs)

    @classmethod
    def _validate_filters(cls, filter_specs, spec, context):
        filters = cls._validate_list_subtypes('filters', Filter, filter_specs, spec, context)
        for filter_spec in filter_specs:
            if filter_spec['type'] == 'facet':
                raise ValidationError(
                    'Target facetting must be declared via the facet field of the target specification, '
                    'specifying filters of type \'facet\' is no longer supported.', spec, 'filters'
                )
        if 'pipeline' in spec:
            raise ValidationError(
                'Target may not declare filters AND a pipeline. Please declare the filters as '
                'part of the pipeline. ', spec, 'filters'
            )
        return filters

    @classmethod
    def _validate_source(cls, source_spec, spec, context):
        if isinstance(source_spec, str):
            if source_spec not in context['sources']:
                msg = f'Target specified non-existent source {source_spec!r}.'
                msg = match_suggestion_message(source_spec, list(context['sources']), msg)
                raise ValidationError(msg, spec, source_spec)
            return source_spec
        warnings.warn(
            'Inlining source definitions in a target is no longer supported. '
            'Please ensure you declare all sources as part of the global \'sources\' '
            'field', DeprecationWarning
        )
        source_spec = Source.validate(source_spec, context)
        src_cls = Source._get_type(source_spec['type'])
        source_name = f'{src_cls.name}' # NOTE: Create unique name
        if 'sources' not in context:
            context['sources'] = {}
        context['sources'][source_name] = source_spec
        return source_name

    @classmethod
    def _validate_views(cls, view_specs, spec, context):
        view_specs = cls._validate_dict_or_list_subtypes('views', View, view_specs, spec, context)
        if 'source' in spec or 'pipeline' in spec:
            return view_specs
        view_specs = view_specs.values() if isinstance(view_specs, dict) else view_specs
        if not all('pipeline' in spec for spec in view_specs):
            raise ValidationError(
                'Target (or its views) must declare a source or a pipeline.', spec
            )
        return view_specs

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def from_spec(cls, spec, **kwargs):
        """
        Creates a Target object from a specification. If a Target
        specification references an existing Source or Filter by name.

        Parameters
        ----------
        spec : dict
          Specification declared as a dictionary of parameter values.
        kwargs: dict
          Additional kwargs to pass to the Target

        Returns
        -------
        Resolved and instantiated Target object
        """
        # Resolve source
        spec = dict(spec)
        views = spec['views']

        pipelines = {}
        source_filters = None
        filter_specs = spec.pop('filters', [])
        view_specs = views if isinstance(views, list) else list(views.values())
        if 'source' in spec:
            source_spec = spec.pop('source', None)
            if isinstance(source_spec, Source):
                source = source_spec
            else:
                source = Source.from_spec(source_spec)
                if isinstance(source_spec, str):
                    source_filters = state.filters.get(source_spec)
        else:
            if 'pipeline' in spec:
                pspecs = [spec['pipeline']]
            else:
                pspecs = [vspec.get('pipeline') for vspec in view_specs if 'pipeline' in vspec]
            pipeline_spec = pspecs[0]
            if isinstance(pipeline_spec, str):
                if pipeline_spec not in state.pipelines:
                    raise ValueError(f'{pipeline_spec!r} not found in global pipelines.')
                pipeline = state.pipelines[pipeline_spec]
            else:
                pipeline = Pipeline.from_spec(pipeline_spec)
            source = pipeline.source
            pipelines[pipeline.table] = pipeline

        tables = source.get_tables()

        # Resolve facets
        if 'facet' in spec:
            facet_spec = spec.pop('facet', {})
            schema = source.get_schema()
        else:
            facet_spec, schema = {}, {}
        spec['facet'] = Facet.from_spec(facet_spec, schema)

        # Backward compatibility
        for view_spec in view_specs:
            if 'pipeline' in view_spec or 'pipeline' in spec:
                continue
            elif 'table' in view_spec:
                table = view_spec['table']
            elif len(tables) == 1:
                table = tables[0]
            else:
                raise ValidationError(
                    'View did not declare unambiguous table reference, '
                    'since selected source declares multiple tables. '
                    'Either declare a pipeline or explicitly declare the '
                    'table that the View references.', view_spec
                )
            pspec = {'table': table}
            if filter_specs:
                pspec['filters'] = filter_specs
            pipelines[table] = Pipeline.from_spec(pspec, source, source_filters)

        # Resolve download options
        download_spec = spec.pop('download', {})
        if 'tables' not in download_spec:
            download_spec['tables'] = list(tables)
        spec['download'] = Download.from_spec(download_spec, pipelines)

        params = dict(kwargs, **spec)
        return cls(source=source, pipelines=pipelines, **params)

    @property
    def refs(self):
        refs = []
        for pipeline in self._pipelines.values():
            for ref in pipeline.refs:
                if ref not in refs:
                    refs.append(ref)
        views = self.views
        for spec in (views if isinstance(views, list) else views.values()):
            for ref in extract_refs(spec, 'variables'):
                if ref not in refs:
                    refs.append(ref)
        return refs

    @property
    def panels(self):
        """
        Returns a layout of the rendered View objects on this target.
        """
        if len(self._cards) > 1:
            default = 'flex' if 'flex' in _LAYOUTS else 'grid'
        else:
            default = 'column'
        layout = self.facet.layout or default
        kwargs = dict(name=self.title, margin=0, sizing_mode='stretch_width')
        if isinstance(layout, dict):
            layout_kwargs = dict(layout)
            layout = layout_kwargs.pop('type')
            kwargs.update(layout_kwargs)
        layout_type = _LAYOUTS[layout]
        if layout == 'grid' and 'ncols' not in kwargs:
            kwargs['ncols'] = 3
        if self._cards:
            content = self._cards
            if len(self._cards) == 1 and not self.show_title:
                self._cards[0]._card.param.set_param(
                    hide_header=True, sizing_mode='stretch_width', margin=0
                )
        else:
            content = (
                pn.pane.Alert(
                    f'**All views are empty for {self.title!r} since no data is available. '
                    'Try relaxing the filter constraints.**',
                    sizing_mode='stretch_width',
                    margin=(0, 0, 50, 0),
                ),
            )
        return layout_type(*content, **kwargs)

    def to_spec(self, context=None):
        spec = super().to_spec(context=context)
        if len(self._pipelines) == 1:
            pipeline = list(self._pipelines.values())[0]
            if context:
                if pipeline.name not in context.get('pipelines', {}):
                    context['pipelines'][pipeline.name] = pipeline.to_spec(context=context)
                spec['pipeline'] = pipeline.name
            else:
                spec['pipeline'] = pipeline.to_spec()
        if 'source' in spec and 'pipeline' in spec:
            del spec['source']
        return spec

    @pn.depends('refresh_rate', watch=True)
    def start(self, event=None):
        """
        Starts any periodic callback instantiated on this object.
        """
        refresh_rate = self.refresh_rate if event is None else event.new
        if refresh_rate is None:
            return
        if self._cb:
            self._cb.period = refresh_rate
        else:
            self._cb = pn.state.add_periodic_callback(
                self.update, refresh_rate
            )

    def update(self, *events, clear_cache=True):
        """
        Updates the views on this target by clearing any caches and
        rerendering the views on this Target.
        """
        if clear_cache:
            self.source.clear_cache()
        self._rerender()
        self._timestamp.object = f'Last updated: {dt.datetime.now().strftime(self.tsformat)}'
